/**************************************/
/*        PROBOJ KSP JESEN 2011       */
/*              DISPLAYER             */
/*               by Imp               */
/**************************************/

#include <cstdio>
#include <cstdlib>
#include <csignal>
#include <stdint.h>
#include <unistd.h>
#include <SDL/SDL.h>
#include <SDL/SDL_ttf.h>
#include <string>
#include <vector>
#include <set>
#include <queue>
#include "../engine/globals.h"
#define WAITTIME 200000

using namespace std;

typedef struct {
    int r, c;
} POINT;

typedef struct {
    uint8_t B;
    uint8_t G;
    uint8_t R;
} COLOR24;

bool operator< (const POINT a, const POINT b) {
    if (a.r == b.r)
        return (a.c < b.c);
    
    return (a.r < b.r);
}

int countries;
bool** neighbours;
int* terrains;
int* earnings;
int* owners;
int* strengths;
POINT* coords;
bool* recruiting;

int players;
vector<string> names;
vector<int> winners;

bool game_running = true;
int current_player = 0;
int current_round = 1;
int money;

SDL_Surface* screen;
SDL_Surface* background;
SDL_Surface *player_icon, *money_icon, *attr_icon;

TTF_Font *bigfont, *smallfont, *hugefont;

char buffer[200];

SDL_Rect player_icon_pos = (SDL_Rect){10, 70, 45, 45};
SDL_Rect player_name_pos = (SDL_Rect){65, 70, 45, 45};
SDL_Rect money_icon_pos = (SDL_Rect){10, 130, 50, 50};
SDL_Rect money_text_pos = (SDL_Rect){65, 130, 50, 50};
SDL_Rect turn_text_pos = (SDL_Rect){10, 10, 50, 50};
SDL_Rect final_text_pos = (SDL_Rect){50, 200, 50, 50};

const COLOR24 colors[] = {(COLOR24){0xB0, 0xB0, 0xB0},
                          (COLOR24){0x00, 0x00, 0xB0},
                          (COLOR24){0x00, 0xB0, 0x00},
                          (COLOR24){0xB0, 0x00, 0x00},
                          (COLOR24){0x00, 0xB0, 0xB0},
                          (COLOR24){0xB0, 0xB0, 0x00}
                         };
                         
const COLOR24 bright_colors[] = {(COLOR24){0xFF, 0xFF, 0xFF},
                                 (COLOR24){0x00, 0x00, 0xFF},
                                 (COLOR24){0x00, 0xFF, 0x00},
                                 (COLOR24){0xFF, 0x40, 0x00},
                                 (COLOR24){0x00, 0xFF, 0xFF},
                                 (COLOR24){0xFF, 0xFF, 0x00}
                                };

const int dr[] = {0, -1, 0, 1, -1, -1,  1, 1,  0, -2, 0, 2, -2, -2,  2, 2,  0, -3, 0, 3};
const int dc[] = {-1, 0, 1, 0, -1,  1, -1, 1, -2,  0, 2, 0, -2,  2, -2, 2, -3,  0, 3, 0};

/* floodfill an area of the screen */
void floodfill(POINT p, COLOR24 color, bool dithered = false, bool borders = false) {
    queue<POINT> Q;
    set<POINT> S;
    Q.push(p);
    
    SDL_LockSurface(background);
    int counter = 0;

    while (!Q.empty()) {
	//if (counter++ > 100) break;
	
        p = Q.front();
        Q.pop();
        
        bool non_border = true;
        
        for (int i = 0; i < 4; i++)
            if ((int)((((COLOR24*)background -> pixels) + ((p.r + dr[i]) * background -> w) + (p.c + dc[i])) -> R) +
                     ((((COLOR24*)background -> pixels) + ((p.r + dr[i]) * background -> w) + (p.c + dc[i])) -> G) +
                     ((((COLOR24*)background -> pixels) + ((p.r + dr[i]) * background -> w) + (p.c + dc[i])) -> B)) {
                    
                    if (S.find((POINT){p.r + dr[i], p.c + dc[i]}) == S.end()) {
                        Q.push((POINT){p.r + dr[i], p.c + dc[i]});
                        S.insert((POINT){p.r + dr[i], p.c + dc[i]});
                    }
            }

        for (int i = 0; i < 20; i++)
            if ((int)((((COLOR24*)background -> pixels) + ((p.r + dr[i]) * background -> w) + (p.c + dc[i])) -> R) +
                     ((((COLOR24*)background -> pixels) + ((p.r + dr[i]) * background -> w) + (p.c + dc[i])) -> G) +
                     ((((COLOR24*)background -> pixels) + ((p.r + dr[i]) * background -> w) + (p.c + dc[i])) -> B))
            {} else non_border = false;
            
        if (!borders || non_border) {
            if (!dithered || (((p.r + p.c) / 5) & 1))
                *(((COLOR24*)(background -> pixels)) + (p.r * background -> w) + p.c) = color;
        } else
            *(((COLOR24*)(background -> pixels)) + (p.r * background -> w) + p.c) = (COLOR24){0, 0, 255};
    }
    SDL_UnlockSurface(background);
}

/* draw the little icons onto the map */
void add_little_images() {
    SDL_BlitSurface(player_icon, NULL, screen, &player_icon_pos);
    
    SDL_LockSurface(screen);
    for (int r = player_icon_pos.y + 5; r < player_icon_pos.y + player_icon_pos.h - 5; r++)
        for (int c = player_icon_pos.x + 5; c < player_icon_pos.x + player_icon_pos.w - 5; c++)
            *(((COLOR24*)(screen -> pixels)) + (r * screen -> w) + c) = colors[current_player];
    SDL_UnlockSurface(screen);
    
    SDL_Surface* text = TTF_RenderText_Blended(bigfont, names[current_player].c_str(), (SDL_Color){0, 0, 0, 255});
    SDL_BlitSurface(text, NULL, screen, &player_name_pos);
    free(text);
    
    SDL_BlitSurface(money_icon, NULL, screen, &money_icon_pos);
    
    sprintf(buffer, "%d", money);
    text = TTF_RenderText_Blended(bigfont, buffer, (SDL_Color){0, 0, 0, 255});
    SDL_BlitSurface(text, NULL, screen, &money_text_pos);
    free(text);
    
    sprintf(buffer, "Kolo %d / %d", current_round, MAXALLOWEDROUNDS);
    text = TTF_RenderText_Blended(bigfont, buffer, (SDL_Color){0, 0, 0, 255});
    SDL_BlitSurface(text, NULL, screen, &turn_text_pos);
    free(text);
    
	return;

    for (int i = 0; i < countries; i++) {
        SDL_Rect img = (SDL_Rect){coords[i].c, coords[i].r, 32, 75};
        SDL_BlitSurface(attr_icon, NULL, screen, &img);
        
        img.x += 40;

        sprintf(buffer, "%d", strengths[i]);
        text = TTF_RenderText_Blended(smallfont, buffer, (SDL_Color){0, 0, 0, 255});
        SDL_BlitSurface(text, NULL, screen, &img);
        free(text);
        
        img.y += 25;

        sprintf(buffer, "%d", terrains[i]);
        text = TTF_RenderText_Blended(smallfont, buffer, (SDL_Color){0, 0, 0, 255});
        SDL_BlitSurface(text, NULL, screen, &img);
        free(text);

        img.y += 25;

        sprintf(buffer, "%d", earnings[i]);
        text = TTF_RenderText_Blended(smallfont, buffer, (SDL_Color){0, 0, 0, 255});
        SDL_BlitSurface(text, NULL, screen, &img);
        free(text);
    }
}

/* signal handler */
void sighandler(int sigtype __attribute__((unused))) {
    exit(2);
}

int main() {
    signal(SIGINT, sighandler);
    
    /* get the map name and prepare the map */
    sprintf(buffer, "cp ../maps/");
    scanf("%50s\n", buffer + 11);
    for (int i = 11; ; i++)
        if (!buffer[i]) {
            sprintf(buffer + i, "/* ../graphics/");
            break;
        }
    system(buffer);
    
    /* get countries count */
    scanf("%d", &countries);
    
    /* get the map */
    neighbours = (bool**)malloc(countries * sizeof(bool*));
    for (int i = 0; i < countries; i++) {
        neighbours[i] = (bool*)malloc(countries * sizeof(bool));
        
        for (int j = 0; j < countries; j++)
            scanf("%d", (int*)&neighbours[i][j]);
    }
    
    /* get the other characteristics */
    terrains = (int*)malloc(countries * sizeof(int));
    earnings = (int*)malloc(countries * sizeof(int));
    owners = (int*)malloc(countries * sizeof(int));
    strengths = (int*)malloc(countries * sizeof(int));
    recruiting = (bool*)malloc(countries * sizeof(bool));
    
    for (int i = 0; i < countries; i++)
        scanf("%d %d %d %d\n", &terrains[i], &earnings[i], &owners[i], &strengths[i]);
    
    /* load the map coordinates */
    coords = (POINT*)malloc(countries * sizeof(POINT));
    FILE* f = fopen("../map/coords", "r");
    for (int i = 0; i < countries; i++)
        fscanf(f, "%d %d", &coords[i].c, &coords[i].r);
    fclose(f);
    
    /* load player names */
    names.push_back("*** Hra sa zacina ***");
    
    scanf("%d\n", &players);
    for (int i = 0; i < players; i++) {
        int buffer_pos;
        for (buffer_pos = 0; buffer_pos < 20; buffer_pos++)
            if ((buffer[buffer_pos] = getchar()) == '\n') break;
        if (buffer[buffer_pos] != '\n')
            while (getchar() != '\n') {};
        buffer[buffer_pos] = '\0';

        names.push_back(buffer);
    }
    
    /* screen initialization */
    SDL_Init(SDL_INIT_VIDEO);
    atexit(SDL_Quit);
    
    screen = SDL_SetVideoMode(1024, 768, 24, SDL_SWSURFACE | SDL_DOUBLEBUF | SDL_NOFRAME | SDL_FULLSCREEN );
    SDL_WM_SetCaption("Proboj KSP 2011", 0);
    
    background = SDL_LoadBMP("../graphics/picture.bmp");
    money_icon = SDL_LoadBMP("../graphics/money.bmp");
    attr_icon = SDL_LoadBMP("../graphics/attributes.bmp");
    player_icon = SDL_LoadBMP("../graphics/player.bmp");
    SDL_SetColorKey(attr_icon, SDL_SRCCOLORKEY, SDL_MapRGB(attr_icon -> format, 255, 255, 255));
    
    /* font initialization */
    TTF_Init();
    atexit(TTF_Quit);
    bigfont = TTF_OpenFont("../fonts/FreeSansBold.ttf", 24);
    smallfont = TTF_OpenFont("../fonts/FreeSansBold.ttf", 16);
    hugefont = TTF_OpenFont("../fonts/FreeSansBold.ttf", 72);

    /* initialize the map */
    for (int i = 0; i < countries; i++)
        floodfill((POINT){coords[i].r, coords[i].c}, colors[owners[i]]);

    SDL_BlitSurface(background, NULL, screen, NULL);
    add_little_images();
    SDL_Flip(screen);
    
    char c;
    int source, target, amount, result;
    
    /* main rendering loop */
    while (game_running) {
        bool won_game = false;
        
        scanf("%c", &c);
        
        switch (c) {
            case 'P':
                scanf("%d %d\n", &current_player, &money);
                
                for (int i = 0; i < countries; i++)
                    recruiting[i] = false;
                
                SDL_BlitSurface(background, NULL, screen, NULL);
                add_little_images();
                SDL_Flip(screen);
            break;
            
            case 'R':
                scanf("%d %d\n", &source, &amount);
                --source;
                
                money -= amount;
                strengths[source] += amount;
                recruiting[source] = true;
            break;
            
            case 'D':
                //scanf("\n");
                
                for (int i = 0; i < countries; i++)
                    if (recruiting[i])
                        floodfill((POINT){coords[i].r, coords[i].c}, bright_colors[owners[i]]);
                
                SDL_BlitSurface(background, NULL, screen, NULL);
                add_little_images();
                SDL_Flip(screen);
		usleep(WAITTIME);
                
                for (int i = 0; i < countries; i++)
                    if (recruiting[i])
                        floodfill((POINT){coords[i].r, coords[i].c}, colors[owners[i]]);
                
                SDL_BlitSurface(background, NULL, screen, NULL);
                add_little_images();
                SDL_Flip(screen);
		usleep(WAITTIME);
            break;
            
            case 'N':
                scanf("\n");
            break;
            
            case 'W':
                won_game = true;
            case 'M': case 'L':
                scanf("%d %d %d %d\n", &source, &target, &amount, &result);
                --source; --target;
                
                strengths[source] -= amount;
                
                for (int k = 0; k < 1; k++) {
                    if (k) {
                        floodfill((POINT){coords[source].r, coords[source].c}, colors[owners[source]]);
                        floodfill((POINT){coords[target].r, coords[target].c}, colors[owners[source]]);
                        SDL_BlitSurface(background, NULL, screen, NULL);
                        add_little_images();
                        SDL_Flip(screen);
			usleep(WAITTIME);
                    }
                    
                    if (owners[source] != owners[target])
                        floodfill((POINT){coords[target].r, coords[target].c}, colors[owners[source]], true);
                    else
                        floodfill((POINT){coords[target].r, coords[target].c}, bright_colors[owners[source]], true);
                    floodfill((POINT){coords[source].r, coords[source].c}, colors[owners[source]], false, true);
                    SDL_BlitSurface(background, NULL, screen, NULL);
                    add_little_images();
                    SDL_Flip(screen);
		    usleep(WAITTIME);
                }
                
                strengths[target] = result;
                
                if (won_game)
                    owners[target] = owners[source];
                
                floodfill((POINT){coords[source].r, coords[source].c}, colors[owners[source]]);
                floodfill((POINT){coords[target].r, coords[target].c}, colors[owners[target]]);
                SDL_BlitSurface(background, NULL, screen, NULL);
                add_little_images();
                SDL_Flip(screen);
            break;
            
            case 'F':
                winners.resize(1);
                scanf("%d", &winners[0]);
                
            case 'E':
                if (!winners.size()) {
                    int x;
                    scanf("%d", &x);
                    winners.resize(x);
                
                    for (int i = 0; i < x; i++)
                        scanf("%d", &winners[i]);
                    //scanf("\n");
                }
                
                game_running = false; {
                    sprintf(buffer, "KONIEC HRY");
                    SDL_Surface* text = TTF_RenderText_Blended(hugefont, buffer, (SDL_Color){0, 0, 0, 255});
                    SDL_BlitSurface(text, NULL, screen, &final_text_pos);
                    free(text);
                    final_text_pos.y += 80;
                    
                    if (winners.size() == 1)
                        sprintf(buffer, "VYHRAL HRAC: ");
                    else
                        sprintf(buffer, "VYHRALI HRACI: ");
                    
                    text = TTF_RenderText_Blended(hugefont, buffer, (SDL_Color){0, 0, 0, 255});
                    SDL_BlitSurface(text, NULL, screen, &final_text_pos);
                    free(text);
                    
                    for (unsigned int i = 0; i < winners.size(); i++) {
                        final_text_pos.y += 80;
                        sprintf(buffer, "%s (%d)", names[winners[i]].c_str(), winners[i]);
                        text = TTF_RenderText_Blended(hugefont, buffer,
                                (SDL_Color){bright_colors[winners[i]].R, 
                                            bright_colors[winners[i]].G, 
                                            bright_colors[winners[i]].B, 255}
                                                     );
                        SDL_BlitSurface(text, NULL, screen, &final_text_pos);
                        free(text);
                    }
                    final_text_pos.y -= 80 * (winners.size() + 1);
                    
                    SDL_Flip(screen);
                }
                
                sleep(5);
                
                putchar('f');
                fflush(stdout);
            break;
            
            case 'S':
                scanf("\n");
                game_running = false;
            break;
            
            case 'B':
                ++current_round;
                putchar('f');
                fflush(stdout);
            break;
        }
        
        
        /* handle keypress */
        SDL_Event event;
        
        while (SDL_PollEvent(&event))
            if (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_c && (event.key.keysym.mod & KMOD_CTRL))
                return 3;
    }
    
    
    return 0;
}
