/**************************************/
/*        PROBOJ KSP JESEN 2011       */
/*             GAME ENGINE            */
/*               by Imp               */
/**************************************/

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <csignal>
#include <ctime>
#include <vector>
#include <set>
#include <algorithm>
#include <unistd.h>
#include <cmath>
#include "globals.h"

using namespace std;

#define IN 0
#define OUT 1
#define PLAYERS 5

FILE* streams[PLAYERS + 1][2];      // pipey na hráčov
FILE* displayer;                    // pipe pre displayer
FILE* dispfeed;                     // pipe pre feedback od displayera
FILE* logger;                       // pipe pre logger

char command[200];                  // buffer pre shellové príkazy

int countries;                      // počet krajín
vector<int> strengths;              // vojenská prítomnosť v krajinách
vector<int> earnings;               // zárobky za krajinu
vector<int> terrains;               // hornatosť krajiny
bool** neighbours;                  // adjacenčná matica
vector<int> owners;                 // majitelia krajín (0 - nikto, x > 0 - hráč x)
int finances[PLAYERS + 1];          // peniaze hráčov
vector<int> ordered;                // poradie hráčov na ťahu
int current_player;                 // momentálny hráč na ťahu (ak nula, vypršal preňho časový limit)
set<int> still_in_game;             // hráči, ktorí nevypadli
vector<int> owned(PLAYERS + 1, 0);  // počet krajín, ktoré hráči vlastnia
bool game_over = false;             // beží stále hra?
int current_round;                  // ktoré kolo prebieha?
char names[PLAYERS + 1][200];       // mena hracov

/* handler pre ctrl+c (SIGINT) */
void sighandler(int sigtype __attribute__((unused))) {
    sprintf(command, "killall -9 hrac");
    system(command);
    
    /* záznam */
    fprintf(logger, "S\n");
    fflush(logger);
    fprintf(displayer, "S\n");
    fflush(displayer);
    
    puts("PROBOJ UKONČENÝ UŽÍVATEĽSKÝM ZÁSAHOM.");
    system("killall -9 displayer");
    system("killall engine");
    exit(2);
}

/* handler pre vypršanie časového limitu (SIGALRM) */
void alarmhandler(int sigtype __attribute__((unused))) {
    puts("Vypršal časový limit, hráč je vylúčený z hry.");
    still_in_game.erase(current_player);
    fclose(streams[current_player][IN]);
    fclose(streams[current_player][OUT]);

    current_player = 0;

    fflush(stdout);
}

/* handler pre zlomenú rúru */
int current_com = 0;

void brokenpipehandler(int sigtype __attribute__((unused))) {
    if (current_com) {
        printf("--Zlomená rúra. Hráč %d je vylúčený z hry.\n", current_com);
    } else {
        puts("--Zlomená rúra s displayerom alebo loggerom.");
        puts("PROBOJ UKONČENÝ");
        fflush(stdout);
        exit(1);
    }
    still_in_game.erase(current_com);
    fflush(stdout);
}

int main(int argc, char* argv[]) {
    puts("Spúšťa sa Proboj KSP Jeseň 2011");
    
    /* nastaviť signal handler pre ctrl+c */
    signal(SIGINT, sighandler);
    signal(SIGALRM, alarmhandler);
    signal(SIGPIPE, brokenpipehandler);
    
    /* dostatok parametrov? */
    if (argc < 2) {
        puts("-- Chýba parameter: názov mapy");
        return 1;
    }
    
    /* nakopírovanie mapy */
    sprintf(command, "cp ../maps/%s/* ../map/ 2>> /dev/null", argv[1]);
    if (system(command)) {
        puts("-- Nedá sa nakopírovať mapa.");
        return 1;
    }
    
    
    /* veľký cyklus, ktorý zakaždým nakopíruje hráčov a spustí hru */
    while (1) {
        /* spustenie hry */
        puts("Začína sa nová hra.");
        
        sprintf(command, "echo === Projekt spustený o `date` === >> ../errorlog/errorlog");
        system(command);
        
        /* inicializácia niektorých globálnych premenných */
        current_com = 0;
        current_round = 0;
        for (int i = 0; i <= PLAYERS; i++)
            owned[i] = 0;
        game_over = false;
        still_in_game.clear();
        strengths.clear();
        earnings.clear();
        owners.clear();
        terrains.clear();
        ordered.clear();
        
        /* zrušiť prípadných bežiacich hráčov */
        printf("Upratanie po poslednom spustení ...");
        fflush(stdout);
        sprintf(command, "killall hrac 2>> ../errorlog/errorlog");
        system(command);
        sprintf(command, "test `killall hrac 2>&1 | wc -l` -eq 0");
        while (!system(command)) {
            putchar('.');
            fflush(stdout);
        }
        sprintf(command, "test `killall displayer 2>&1 | wc -l` -eq 0");
        while (!system(command)) {
            putchar('.');
            fflush(stdout);
        }
        puts(" hotovo.");

        /* načítanie mapy */
        
        puts("Načítavam mapu.");
        
        FILE* f = fopen("../map/map", "r");
        fscanf(f, "%d", &countries);
        
        neighbours = (bool**)malloc(countries * sizeof(bool*));
        for (int i = 0; i < countries; i++)
            neighbours[i] = (bool*)malloc(countries * sizeof(bool));
        
        for (int i = 0; i < countries; i++)
            for (int j = 0; j < countries; j++)
                neighbours[i][j] = false;
        
        for (int i = 0; i < countries; i++) {
            int neighbour_count;
            fscanf(f, "%d", &neighbour_count);
            
            for (int j = 0; j < neighbour_count; j++) {
                int x;
                
                fscanf(f, "%d", &x); --x;
                neighbours[i][x] = neighbours[x][i] = true;
            }
        }
        
        fclose(f);
        
        /* nacitanie terénu */
        
        puts("Načítavam terén.");
        f = fopen("../map/terrain", "r");
        
        for (int i = 0; i < countries; i++) {
            int x;
            
            fscanf(f, "%d", &x);
            terrains.push_back(x);
            fscanf(f, "%d", &x);
            earnings.push_back(x);
        }
        
        fclose(f);
        
        /* načítanie vojsk */
        
        puts("Načítavam vojská.");
        f = fopen("../map/armies", "r");
        
        for (int i = 0; i < countries; i++) {
            int x;
            
            fscanf(f, "%d", &x);
            owners.push_back(x);
            fscanf(f, "%d", &x);
            strengths.push_back(x);
        }
        
        fclose(f);
        
        /* náhodné prepermutovanie počiatočného rozostavenia hráčov */
        puts("Generujem náhodné rozostavenie.");
        
        srand(time(NULL));
        f = fopen("../map/currentarmies", "w");
        
        for (int i = 0; i <= PLAYERS; i++) ordered.push_back(i);
        random_shuffle(ordered.begin() + 1, ordered.end());
	for (int i = 0; i <= PLAYERS; i++)
		if (ordered[i] == 5) {
			ordered[i] = ordered[5];
			ordered[5] = 5;
			break;
		}
        
        for (int i = 0; i < countries; i++) {
            fprintf(f, "%d %d\n", owners[i] = ordered[owners[i]], strengths[i]);
            ++owned[owners[i]];
        }
            
        fclose(f);
        
        /* náhodné určenie poradia na ťahu*/
        printf("Žrebujem poradie hráčov na ťahu: ");
        
        random_shuffle(ordered.begin() + 1, ordered.end());
        
    	f = fopen("../map/playerorder", "w");
	    for (int i = 1; i <= PLAYERS; i++)
	    	fprintf(f, "%d ", ordered[i]);
	    fclose(f);
	    for (int i = 1; i <= PLAYERS; i++) {
	    	sprintf(command, "cp ../map/playerorder ../prepared/player%d/playerorder", i);
	    	system(command);
	    }        
        
        for (int i = 1; i <= PLAYERS; i++)
            printf("%d ", ordered[i]);
        putchar('\n');
        
        /* vytváram logy */
        puts("Vytváram súbor pre záznam z hry.");
        time_t current_raw_time = time(NULL);
        struct tm* current_time = localtime(&current_raw_time);
                
        sprintf(command, "../logs/proboj-log-%d-%d-%02d-%02d-%02d", current_time -> tm_mday, current_time -> tm_mon,
                current_time -> tm_hour, current_time -> tm_min, current_time -> tm_sec);
        logger = fopen(command, "w");
        
        /* pripravenie displayera */
        printf("Pripravujem displayer...  ");
        fflush(stdout);

        if (!system("../displayer/displayer < ../displayer/displayer.out > ../displayer/displayer.in &")) {
            displayer = fopen("../displayer/displayer.out", "w");
            dispfeed = fopen("../displayer/displayer.in", "r");
            puts("OK");
        } else {
            puts("CHYBA!");
            puts("Nedá sa spustiť displayer.");
            return 1;
        }
        
        /* poslanie informácií o mape */
        printf("Posielam informácie o mape...  ");
        fprintf(logger, "%s\n", argv[1]);
        fprintf(displayer, "%s\n", argv[1]);
        fprintf(logger, "%d\n", countries);
        fprintf(displayer, "%d\n", countries);
        
        for (int i = 0; i < countries; i++) {
            for (int j = 0; j < countries; j++) {
                fprintf(logger, "%d ", (int)neighbours[i][j]);
                fprintf(displayer, "%d ", (int)neighbours[i][j]);
            }
            
            fprintf(logger, "\n");
            fprintf(displayer, "\n");
        }
        
        for (int i = 0; i < countries; i++) {
            fprintf(logger, "%d %d %d %d\n", terrains[i], earnings[i], owners[i], strengths[i]);
            fprintf(displayer, "%d %d %d %d\n", terrains[i], earnings[i], owners[i], strengths[i]);
        }
        
        puts("OK");
        
        /* nakopírovanie a spustenie hráčov */
        puts("Pripravujem hráčov...");
        
        for (int i = 1; i <= PLAYERS; i++) {
            
            sprintf(command, "cp ../compiled/player%d/hrac ../prepared/player%d/hrac 2>> ../errorlog/errorlog", i, i);
            if (system(command)) {
                printf("  Chýba hráč č. %d !\nPROBOJ SA VYPÍNA. ZAVOLAJTE ADMINISTRÁTORA.\n", i);
                return 1;
            } else {
                int ret = 0;
                
                sprintf(command, "cp ../map/map ../prepared/player%d/mapa 2>> ../errorlog/errorlog", i);
                ret |= system(command);
                sprintf(command, "cp ../map/terrain ../prepared/player%d/teren 2>> ../errorlog/errorlog", i);
                ret |= system(command);
                sprintf(command, "cp ../map/currentarmies ../prepared/player%d/vojska 2>> ../errorlog/errorlog", i);
                ret |= system(command);
                sprintf(command, "chown player%d:proboj ../prepared/player%d 2>> ../errorlog/errorlog", i, i);
                ret |= system(command);
                sprintf(command, "chown player%d:proboj ../prepared/player%d/* 2>> ../errorlog/errorlog", i, i);
                ret |= system(command);
                sprintf(command, "chmod a+x,u+s ../prepared/player%d/hrac 2>> ../errorlog/errorlog", i);
                ret |= system(command);
                sprintf(command, "../prepared/player%d/hrac > ../pipes/%d.in < ../pipes/%d.out &", i, i, i);
                ret |= system(command);
                
                if (ret) {
                    printf("  Problém s hráčom č. %d !\nPROBOJ SA VYPÍNA. ZAVOLAJTE ADMINISTRÁTORA.\n", i);
                    return 1;
                } else printf("  Hráč č. %d pripravený.\n", i);
            }
            
            sprintf(command, "../pipes/%d.in", i);
            streams[i][IN] = fopen(command, "r");
            sprintf(command, "../pipes/%d.out", i);
            streams[i][OUT] = fopen(command, "w");
            
            still_in_game.insert(i);
        }
        
        /* spustenie hry */
        
        puts("Čakám na prihlásenie hráčov.");
        
        fprintf(displayer, "%d\n", PLAYERS);
        fprintf(logger, "%d\n", PLAYERS);
        
        for (int i = 1; i <= PLAYERS; i++) {
            fprintf(streams[i][OUT], "%d\n", i);
            fflush(streams[i][OUT]);
            
            printf("  Prihlásenie hráča %d ... ", i);
            fflush(stdout);
            
            int counter;
            for (counter = 0; counter < 20; counter++)
                if ((command[counter] = fgetc(streams[i][IN])) == '\n') break;
            if (command[counter] != '\n')
                while (fgetc(streams[i][IN]) != '\n') {};
            command[counter] = '\0';
            
            fprintf(displayer, "%s\n", command);
            fprintf(logger, "%s\n", command);
	    strcpy(names[i], command);
	    puts(names[i]);
            
            finances[i] = 0;
        }
        
        /* beh hry */
        puts("BEŽÍ HRA ...");
        
        while (1) {
            
            /* je čas na ďalšie kolo? ak nie, vyhlás víťazov */
            if (++current_round > MAXALLOWEDROUNDS) {
                vector<int> winners;
                int max_owned_countries = -1;
                
                for (int i = 1; i <= PLAYERS; i++)
                    if (owned[i] > max_owned_countries)
                        max_owned_countries = owned[i];
                
                for (int i = 1; i <= PLAYERS; i++)
                    if (owned[i] == max_owned_countries)
                        winners.push_back(i);
                
                fprintf(displayer, "E");
                fprintf(displayer, " %u", winners.size());
                for (unsigned int i = 0; i < winners.size(); i++)
                    fprintf(displayer, " %d", winners[i]);
                fprintf(displayer, "\n");
                fflush(displayer);
                
                /* záznam */
                fprintf(logger, "E");
                fprintf(logger, " %u", winners.size());
                for (unsigned int i = 0; i < winners.size(); i++)
                    fprintf(logger, " %d", winners[i]);
                fprintf(logger, "\n");
                fflush(logger);
                
                printf("Skončil sa povolený počet ťahov.\n");
                printf("\nČakám na displayer... ");
                fflush(stdout);
                *command = fgetc(dispfeed);
                puts("hotovo.");
                break;
            }
            
            /* pre každého hráča, ako idú na ťahu... */
            for (int i = 1; i <= PLAYERS; i++) {
                /* je ešte v hre? */
                if (!still_in_game.size()) {
                    printf("******* VŠETCI HRÁČI VYPADLI Z HRY!*******\n");
                    game_over = true;
                    break;
                }
                
                if (still_in_game.find(ordered[i]) == still_in_game.end()) continue;
                
                if (still_in_game.size() == 1) {
                    printf("********** HRÁČ %d VYHRAL! **********\n", ordered[i]);
                    
                    /* záznam */
                    fprintf(logger, "F %d\n", *still_in_game.begin());
                    fflush(logger);
                    fprintf(displayer, "F %d\n", *still_in_game.begin());
                    fflush(displayer);
                    
                    printf("\nČakám na displayer... ");
                    fflush(stdout);
                    *command = fgetc(dispfeed);
                    puts("hotovo.");
                    
                    game_over = true;
                    break;
                }
                //sleep(1);
                
                
                /* prirátanie a oznámenie financií */
                for (int j = 0; j < countries; j++)
                    if (owners[j] == ordered[i])
                        finances[ordered[i]] += earnings[j];
                
                printf("Na ťahu je hráč %d.\n  Ovláda %d krajín.\n  Má %d zlatiek.\n",
                       ordered[i], owned[ordered[i]], finances[ordered[i]]);
                
                fprintf(streams[ordered[i]][OUT], "%d %d\n", finances[ordered[i]], owned[ordered[i]]);
                fflush(streams[ordered[i]][OUT]);
                
                /* záznam */
                fprintf(logger, "P %d %d\n", ordered[i], finances[ordered[i]]);
                fflush(logger);
                fprintf(displayer, "P %d %d\n", ordered[i], finances[ordered[i]]);
                fflush(displayer);
                
                /* príprava na komunikáciu s hráčom */
                current_player = ordered[i];
                alarm(2);
                
                /* najímanie vojakov */
                while (1) {
                    int country_id = 0, soldiers_count = 0;
                    fscanf(streams[ordered[i]][IN], "%d %d", &country_id, &soldiers_count);
                    
                    if (!country_id--) break;
                    if (!current_player) break;
                    
                    /* správny rozkaz? */
                    if (country_id >= 0 && country_id < countries && soldiers_count >= 0 &&
                        owners[country_id] == ordered[i] && soldiers_count <= finances[ordered[i]]) {
                        
                        strengths[country_id] += soldiers_count;
                        finances[ordered[i]] -= soldiers_count;
                        
                        fprintf(streams[ordered[i]][OUT], "1\n");
                        fflush(streams[ordered[i]][OUT]);
                        
                        printf("  Najíma %d vojakov v %d, zostalo mu %d zlatiek.\n",
                               soldiers_count, country_id + 1, finances[ordered[i]]);
                        
                        current_com = 0;
                        for (int j = 1; j <= PLAYERS; j++)
                            if (j != ordered[i] && still_in_game.find(j) != still_in_game.end()) {
                                current_com = j;
                                fprintf(streams[j][OUT], " -3 %d %d\n", country_id + 1, soldiers_count);
                                fflush(streams[j][OUT]);
                            }
                            
                        /* záznam */
                        fprintf(logger, "R %d %d\n", country_id + 1, soldiers_count);
                        fflush(logger);
                        fprintf(displayer, "R %d %d\n", country_id + 1, soldiers_count);
                        fflush(displayer);

                    } else {
                        puts("  Chybný príkaz na najímanie.");
                        fprintf(streams[ordered[i]][OUT], "0\n");
                        fflush(streams[ordered[i]][OUT]);
                    }
                    
                    fflush(stdout);
                }
                
                /* záznam */
                fprintf(logger, "D\n");
                fflush(logger);
                fprintf(displayer, "D\n");
                fflush(displayer);
                
                /* presuny a útoky */
                if (current_player) {
                    int will_attack = 0, source = 0, target = 0, amount = 0;
                    fscanf(streams[ordered[i]][IN], "%d", &will_attack);
                   
                    /* hýbe sa v tomto ťahu? */
                    if (will_attack) {
                        fscanf(streams[ordered[i]][IN], "%d %d %d", &source, &target, &amount);
                        --source; --target;
                        
                        /* správny rozkaz? */
                        if (source >= 0 && source < countries && target >= 0 && target < countries &&
                            amount > 0 && amount <= strengths[source] && owners[source] == ordered[i] &&
                            neighbours[source][target] && source != target) {
                            
                                /* presúva? */
                                if (owners[target] == ordered[i]) {
                                    printf("  Presúva %d (z %d) vojakov z %d do %d\n",
                                           amount, strengths[source], source + 1, target + 1);
                                    
                                    strengths[source] -= amount;
                                    strengths[target] += amount;
                                    
                                    fprintf(streams[ordered[i]][OUT], "2 %d\n", strengths[target]);
                                    fflush(streams[ordered[i]][OUT]);
                                    
                                    current_com = 0;
                                    for (int j = 1; j <= PLAYERS; j++)
                                        if (j != ordered[i] && still_in_game.find(j) != still_in_game.end()) {
                                            current_com = j;
                                            fprintf(streams[j][OUT], " -1 %d %d %d %d\n",
                                                    source + 1, target + 1, amount, strengths[target]);
                                            fflush(streams[j][OUT]);
                                        }
                                
                                    /* záznam */
                                    fprintf(logger, "M %d %d %d %d\n", source + 1, target + 1,
                                                                       amount, strengths[target]);
                                    fflush(logger);
                                    fprintf(displayer, "M %d %d %d %d\n", source + 1, target + 1,
                                                                          amount, strengths[target]);
                                    fflush(displayer);
                                    
                                /* útočí? */
                                } else {
                                    
                                    strengths[source] -= amount;
                            
                                    double attacker = amount;
                                    double defender = strengths[target];
                                    double handicap = pow(0.8, terrains[target]);
                                    const double granularity = 0.001;
                                    const double dispersion = 0.1;

                                    double ra = (1 - dispersion) + (1.0 * rand() / RAND_MAX) * (2 * dispersion);
                                    double rd = (1 - dispersion) + (1.0 * rand() / RAND_MAX) * (2 * dispersion);
                                    
                                    while (attacker > 0 && defender > 0) {
                                        double newa = attacker - (sqrt(defender) * granularity * rd);
                                        double newd = defender - (sqrt(attacker) * granularity * ra * handicap);
                                        
                                        attacker = newa;
                                        defender = newd;
                                    }
                                
                                    /* vyhral?*/
                                    if (attacker > defender) {
                                        ++owned[owners[source]];
                                        --owned[owners[target]];

                                        if (!(strengths[target] = round(attacker)))
                                            strengths[target] = 1;
                                        
                                        
                                        printf("  Vyhráva, prežilo %d vojakov z %d.\n", strengths[target], amount);
                                        
                                        current_com = 0;
                                        for (int j = 1; j <= PLAYERS; j++)
                                            if (j != ordered[i] && still_in_game.find(j) != still_in_game.end()) {
                                                current_com = j;
                                                fprintf(streams[j][OUT], " -1 %d %d %d %d\n",
                                                        source + 1, target + 1, amount, strengths[target]);
                                                fflush(streams[j][OUT]);
                                            }
                                        
                                        /* záznam */
                                        fprintf(logger, "W %d %d %d %d\n", source + 1, target + 1,
                                                                           amount, strengths[target]);
                                        fflush(logger);
                                        fprintf(displayer, "W %d %d %d %d\n", source + 1, target + 1,
                                                                              amount, strengths[target]);
                                        fflush(displayer);
                                            
                                        if (owners[target] && !owned[owners[target]]) {
                                            still_in_game.erase(owners[target]);
                                            printf("*** HRÁČ %d JE PORAZENÝ! ***\n", owners[target]);
                                            printf("*** V hre zostávajú: ");
                                            for (set<int>::iterator iter = still_in_game.begin();
                                                 iter != still_in_game.end(); iter++)
                                                    printf("%d ", *iter);
                                            puts("***");
                                            fprintf(streams[owners[target]][OUT], "0 0\n");
                                            fflush(streams[owners[target]][OUT]);
                                        }
                                        
                                        owners[target] = ordered[i];
                                    } else {
                                        int oldstrength = strengths[target];
                                        strengths[target] = round(defender);
                                    
                                        printf("  Prehráva, súperovi zostáva %d vojakov z %d.\n",
                                               strengths[target], oldstrength);
                                        
                                       
                                        
                                        current_com = 0;
                                        for (int j = 1; j <= PLAYERS; j++)
                                            if (j != ordered[i] && still_in_game.find(j) != still_in_game.end()) {
                                                current_com = j;
                                                fprintf(streams[j][OUT], " -2 %d %d %d %d\n",
                                                        source + 1, target + 1, amount, strengths[target]);
                                                fflush(streams[j][OUT]);
                                            }
                                            
                                        /* záznam */
                                        fprintf(logger, "L %d %d %d %d\n", source + 1, target + 1,
                                                                           amount, strengths[target]);
                                        fflush(logger);
                                        fprintf(displayer, "L %d %d %d %d\n", source + 1, target + 1,
                                                                              amount, strengths[target]);
                                        fflush(displayer);
                                    }
                                
                                    fprintf(streams[ordered[i]][OUT], "%d %d\n",
                                            (int)(attacker > defender) + 1, strengths[target]);
                                    fflush(streams[ordered[i]][OUT]);
                                }
                                
                        } else {
                            puts("  Chybný príkaz na presun / útok.");
                        
                            fprintf(streams[ordered[i]][OUT], "0 %d\n", strengths[target]);
                            fflush(streams[ordered[i]][OUT]);
                            
                            /* záznam */
                            fprintf(logger, "N\n");
                            fflush(logger);
                            fprintf(displayer, "N\n");
                            fflush(displayer);
                        }
                    } else {
                        puts("  Neútočí.");
                        
                        /* záznam */
                        fprintf(logger, "N\n");
                        fflush(logger);
                        fprintf(displayer, "N\n");
                        fflush(displayer);
                    }
                        
                    fflush(stdout);
                }
                
                /* resetovanie časovača */
		alarm(0);
            }
            
            /* nová hra? */
            if (game_over) break;
            
            /* koniec kola, čakanie na odpoveď displayera */
            current_com = 0;
            printf("\nKoniec kola.\n");
            fflush(stdout);
            if (argc < 3 || strcmp(argv[2], "-n")) {
                
                fprintf(displayer, "B\n");
                fflush(displayer);
                printf("Čakám na vykreslenie... ");
                fflush(stdout);
                *command = fgetc(dispfeed);
                puts("hotovo\n");
            }
            puts("");
            fflush(stdout);
        }
        
        /* vyčistenie pamäte pred ďalším kolom */
        for (int i = 0; i < countries; i++)
            free(neighbours[i]);
        free(neighbours);

	for (int i = 1; i <= PLAYERS; i++) {
		fclose(streams[i][IN]);
		fclose(streams[i][OUT]);
	}
	system("killall -9 hrac");
    }
    
    return 0;
}
